@@ -30,9 +30,15 @@
 
 #include "cupsconnection.h"
 #include "cupsppd.h"
+#include "cupsipp.h"
 
-static PyObject *cups_password_callback = NULL;
+PyObject *cups_password_callback = NULL;
+
+#ifdef HAVE_CUPS_1_4
+PyObject *cups_password_callback_context = NULL;
+#else /* !HAVE_CUPS_1_4 */
 void *g_current_connection = NULL;
+#endif /* !HAVE_CUPS_1_4 */
 
 //////////////////////
 // Worker functions //
@@ -101,6 +107,7 @@ do_model_compare (const wchar_t *a, const wchar_t *b)
   return 1;
 }
 
+#ifndef HAVE_CUPS_1_4
 static const char *
 do_password_callback (const char *prompt)
 {
@@ -141,6 +148,7 @@ do_password_callback (const char *prompt)
   debugprintf ("<- do_password_callback\n");
   return password;
 }
+#endif /* !HAVE_CUPS_1_4 */
 
 //////////////////////////
 // Module-level methods //
@@ -287,6 +295,7 @@ cups_getEncryption (PyObject *self)
 static PyObject *
 cups_setPasswordCB (PyObject *self, PyObject *args)
 {
+  static PyObject *current_cb_context;
   PyObject *cb;
 
   if (!PyArg_ParseTuple (args, "O:cups_setPasswordCB", &cb))
@@ -297,14 +306,71 @@ cups_setPasswordCB (PyObject *self, PyObject *args)
     return NULL;
   }
 
+  debugprintf ("-> cups_setPasswordCB\n");
+  Py_XDECREF (current_cb_context);
+  current_cb_context = NULL;
+
   Py_XINCREF (cb);
   Py_XDECREF (cups_password_callback);
   cups_password_callback = cb;
+
+#ifdef HAVE_CUPS_1_4
+  cupsSetPasswordCB2 (password_callback_oldstyle, NULL);
+#else
   cupsSetPasswordCB (do_password_callback);
+#endif
+
+  debugprintf ("<- cups_setPasswordCB\n");
   Py_INCREF (Py_None);
   return Py_None;
 }
 
+#ifdef HAVE_CUPS_1_4
+static PyObject *
+cups_setPasswordCB2 (PyObject *self, PyObject *args)
+{
+  static PyObject *current_cb_context;
+  PyObject *cb;
+  PyObject *cb_context = NULL;
+
+  if (!PyArg_ParseTuple (args, "O|O", &cb, &cb_context))
+    return NULL;
+
+  if (cb == Py_None && cb_context != NULL) {
+    PyErr_SetString (PyExc_TypeError, "Default callback takes no context");
+    return NULL;
+  }
+  else if (cb != Py_None && !PyCallable_Check (cb)) {
+    PyErr_SetString (PyExc_TypeError, "Parameter must be callable");
+    return NULL;
+  }
+
+  debugprintf ("-> cups_setPasswordCB2\n");
+
+  Py_XINCREF (cb_context);
+  Py_XDECREF (current_cb_context);
+  current_cb_context = cb_context;
+
+  if (cb == Py_None)
+  {
+    Py_XDECREF (cups_password_callback);
+    cups_password_callback = NULL;
+    cupsSetPasswordCB2 (NULL, NULL);
+  }
+  else
+  {
+    Py_XINCREF (cb);
+    Py_XDECREF (cups_password_callback);
+    cups_password_callback = cb;
+    cupsSetPasswordCB2 (password_callback_newstyle, cb_context);
+  }
+
+  debugprintf ("<- cups_setPasswordCB2\n");
+  Py_INCREF (Py_None);
+  return Py_None;
+}
+#endif /* HAVE_CUPS_1_4 */
+
 static PyObject *
 cups_ppdSetConformance (PyObject *self, PyObject *args)
 {
@@ -420,6 +486,19 @@ static PyMethodDef CupsMethods[] = {
     "@type fn: callable object\n"
     "@param fn: callback function" },
 
+#ifdef HAVE_CUPS_1_4
+  { "setPasswordCB2", cups_setPasswordCB2, METH_VARARGS,
+    "setPasswordCB2(fn, context=None) -> None\n\n"
+    "Set password callback function.  This Python function will be called \n"
+    "when a password is required.  It must take parameters of type string \n"
+    "(the password prompt), instance (the cups.Connection), string (the HTTP "
+    "method), string (the HTTP resource) and, optionally, the user-supplied "
+    "context.  It must return a string (the password).  To \n"
+    "abort the operation it may return the empty string ('').\n\n"
+    "@type fn: callable object, or None for default handler\n"
+    "@param fn: callback function" },
+#endif /* HAVE_CUPS_1_4 */
+
   { "ppdSetConformance", cups_ppdSetConformance, METH_VARARGS,
     "ppdSetConformance(level) -> None\n\n"
     "Set PPD conformance level.\n\n"
@@ -499,6 +578,22 @@ initcups (void)
   PyModule_AddObject (m, "Dest",
 		      (PyObject *)&cups_DestType);
 
+  // IPPRequest type
+  cups_IPPRequestType.tp_new = PyType_GenericNew;
+  if (PyType_Ready (&cups_IPPRequestType) < 0)
+      return;
+
+  PyModule_AddObject (m, "IPPRequest",
+		      (PyObject *)&cups_IPPRequestType);
+
+  // IPPAttribute type
+  cups_IPPAttributeType.tp_new = PyType_GenericNew;
+  if (PyType_Ready (&cups_IPPAttributeType) < 0)
+      return;
+
+  PyModule_AddObject (m, "IPPAttribute",
+		      (PyObject *)&cups_IPPAttributeType);
+
   // Constants
 
 #define INT_CONSTANT(name)					\
@@ -649,6 +744,31 @@ initcups (void)
   INT_CONSTANT (IPP_MULTIPLE_JOBS_NOT_SUPPORTED);
   INT_CONSTANT (IPP_PRINTER_IS_DEACTIVATED);
 
+  // IPP states
+  INT_CONSTANT (IPP_ERROR);
+  INT_CONSTANT (IPP_IDLE);
+  INT_CONSTANT (IPP_HEADER);
+  INT_CONSTANT (IPP_ATTRIBUTE);
+  INT_CONSTANT (IPP_DATA);
+
+  // IPP attribute tags
+  INT_CONSTANT (IPP_TAG_ZERO);
+  INT_CONSTANT (IPP_TAG_OPERATION);
+  INT_CONSTANT (IPP_TAG_JOB);
+  INT_CONSTANT (IPP_TAG_PRINTER);
+  INT_CONSTANT (IPP_TAG_INTEGER);
+  INT_CONSTANT (IPP_TAG_BOOLEAN);
+  INT_CONSTANT (IPP_TAG_ENUM);
+  INT_CONSTANT (IPP_TAG_STRING);
+  INT_CONSTANT (IPP_TAG_RANGE);
+  INT_CONSTANT (IPP_TAG_TEXT);
+  INT_CONSTANT (IPP_TAG_NAME);
+  INT_CONSTANT (IPP_TAG_KEYWORD);
+  INT_CONSTANT (IPP_TAG_URI);
+  INT_CONSTANT (IPP_TAG_CHARSET);
+  INT_CONSTANT (IPP_TAG_LANGUAGE);
+  INT_CONSTANT (IPP_TAG_MIMETYPE);
+
   // Limits
   INT_CONSTANT (IPP_MAX_NAME);
 
